/*
 * File access convenience functions.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>

#include <string.h>

#include "dataspace_client.h"
#include "file_client.h"
#include "opener_client.h"
#include "opener_context_client.h"
#include "mapped_file_client.h"

#include "file.h"



/* Release resources for the given file. */

void file_close(file_t *file)
{
  if (l4_is_valid_cap(file->ref))
    ipc_cap_free_um(file->ref);

  if (file->memory != NULL)
    ipc_detach_dataspace(file->memory);

  file_init(file);
}

/* Initialise a file structure for a context obtained from the given server. */

long file_context(file_t *file, l4_cap_idx_t server)
{
  client_Opener opener(server);
  unsigned long size, flags;
  long err;

  file_init(file);

  err = opener.context(&file->ref);
  if (err)
    return err;

  client_Dataspace context_ds(file->ref);

  err = context_ds.info(&size, &flags);
  if (err)
    return err;

  file->start_pos = 0;
  file->end_pos = size;

  return ipc_attach_dataspace(file->ref, size, (void **) &file->memory);
}

/* Open a file using the given structure and context. */

long file_context_open(file_t *file, file_t *context)
{
  client_OpenerContext openercontext(context->ref);
  file_init(file);
  return openercontext.open(L4_FPAGE_RW, &file->size, &file->ref);
}

/* Initialise the given file structure. */

void file_init(file_t *file)
{
  file->memory = NULL;
  file->ref = L4_INVALID_CAP;
  file->start_pos = 0;
  file->end_pos = 0;
  file->data_end = 0;
}

/* Open a file using the given structure, indicating the filename and
   filesystem server. This is a convenience function invoking file_context and
   file_context_open. */

long file_open(file_t *file, const char *filename, l4_cap_idx_t server)
{
  file_t context;
  long err;

  err = file_context(&context, server);
  if (err)
    return err;

  if (!file_string_set(&context, filename, 0, NULL))
    return -L4_ENOMEM;

  err = file_context_open(file, &context);
  file_close(&context);
  return err;
}



/* Map a region of the given file to a memory region. */

long file_mmap(file_t *file, offset_t position, offset_t length)
{
  client_MappedFile mapped_file(file->ref);
  long err = mapped_file.mmap(position, length, &file->start_pos, &file->end_pos, &file->data_end);

  if (err)
    return err;

  return ipc_attach_dataspace(file->ref, file_span(file), (void **) &file->memory);
}

/* Resize a file. */

long file_resize(file_t *file, offset_t size)
{
  client_File _file(file->ref);
  offset_t file_size = size;
  long err = _file.resize(&file_size);

  if (!err)
  {
    /* Determine the extent of the file in this region. */

    if (file_size > file->end_pos)
      file->data_end = file_span(file);
    else
      file->data_end = file_size - file->start_pos;

    /* Update the file size locally. */

    file->size = file_size;
  }

  return err;
}



/* Return the amount of data in the mapped region for the given file. */

offset_t file_populated_span(file_t *file)
{
  offset_t size = file_span(file);
  return (file->data_end < size) ? file->data_end : size;
}

/* Return the size of the mapped region for the given file. */

offset_t file_span(file_t *file)
{
  return file->end_pos - file->start_pos;
}



/* Get a pointer to any terminated string at the given offset or NULL if the
   data from offset is not terminated. */

char *file_string_get(file_t *file, offset_t offset)
{
  offset_t limit = file_span(file) - offset;

  if (strnlen(file->memory + offset, limit) < limit)
    return file->memory + offset;
  else
    return NULL;
}

/* Copy a string to the mapped region at the given offset, returning 1 (true)
   where all characters were copied, 0 (false) otherwise. The precise number of
   characters copied, excluding the zero terminator is provided via the written
   parameter if it is not specified as NULL. */

int file_string_set(file_t *file, const char *data, offset_t offset,
                    offset_t *written)
{
  offset_t i, pos, limit = file_span(file);

  /* Do not attempt to copy data with an invalid offset. */

  if (offset >= limit)
  {
    if (written != NULL)
      *written = 0;
    return 0;
  }

  /* Copy the data to the given offset, stopping at the end of the region. */

  for (i = 0, pos = offset; pos < limit; i++, pos++)
  {
    file->memory[pos] = data[i];

    /* Terminator written, can return immediately. */

    if (!data[i])
    {
      if (written != NULL)
        *written = pos - offset;
      return 1;
    }
  }

  /* Terminate the incomplete string at the end of the region. */

  file->memory[limit - 1] = '\0';
  if (written != NULL)
    *written = limit - 1 - offset;
  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
